import numpy as np
from numpy.fft import fft2, ifft2
import matplotlib.pyplot as plt
import skimage.io as io
from numpy.random import poisson
from numpy.random import random_sample
from pypher.pypher import psf2otf
from pathlib import Path
from textwrap import wrap
from scipy.special import gammaln
import os
import sys

from fspecial import fspecial_gaussian_2d
from first_diff import opDx
from first_diff import opDtx
from RLTV_utils import compute_grad_mag, plot_grad_mag
from RLTV_utils import plot_hist_to_screen, plot_grad_hist_to_screen
from RLTV_utils import plot_grad_mag_seq, plot_grad_hist_to_file
from RLTV_utils import compute_gradient_std_dev, plot_hist_to_file
from RLTV_utils import compute_pct_under_threshold, check_create_dir
from RLTV_utils import compute_ground_truth_stats
from RLTV import run_RLTV

def compute_psnr(x_true, x_recon, n_photons):
    mse = np.mean((x_true - x_recon) ** 2)
    psnr = 10 * np.log10(n_photons ** 2 / mse)
    return psnr

# uncomment one of the two below lines to select the default regularization term
# for the RL+TV prior x update rule
#
# RL+TV_A is the x update rule from the original EE367 Lecture 10 notes, Eq.27
# downloaded on 19 February 2021
#
# RL+TV_B is the x update rule from the updated EE367 Lecture 10 notes, Eq. 28
# downloaded on 22 February 2021
#
# use_reg_term = 'RL+TV_A'
use_reg_term = 'RL+TV_B'

filenum = 1

if __name__ == "__main__":
    for i, arg in enumerate(sys.argv):
        if arg.isnumeric():
            filenum = int(arg)
            if (filenum <= 0) or (filenum > 7):
                filenum = 1
        if (arg == "A") or (arg == "B"):
            use_reg_term = "RL+TV_" + arg

print("Using " + use_reg_term) 
print("Using image file " + str(filenum))

image_names = [("FluorescentCells.jpg", "cells"),
               ("hubble_deep_field_1.png", "hubble"),
               ("art_rby_1.png", "art"),
               ("trv-res_1.png", "testpat"),
               ("coffee.png", "coffee"),
               ("motorcycle_right.png", "motorcycle"),
               ("bluecells.png", "cells_blue")]

img_pathname = 'test_images/' + image_names[filenum - 1][0]
img_str = image_names[filenum - 1][1]
print("Using " + img_pathname)

outdir = "./" + img_str + "_" + use_reg_term
check_create_dir(outdir)
check_create_dir(outdir + "/grad_mag_dir")
check_create_dir(outdir + "/grad_hist_dir")
check_create_dir(outdir + "/update_rule_dir")

img = io.imread(img_pathname).astype(np.float32)/255
height, width = img.shape[0], img.shape[1]
if img.shape[2] > 3:
    # image has an alpha channel
    # in the set of test images, all images are opaque (alpha = 1.0)
    # therefore, just ignore alpha
    img = img[:,:,0:3]
x_gt = img.transpose(2, 0, 1)

# PSF and OTF
psf = fspecial_gaussian_2d((15, 15), 1.5)
otf = psf2otf(psf, (height, width))

# Number of photons
n_photons = 100
x_gt = x_gt * n_photons
x_gt_grad_mag = compute_grad_mag(x_gt)
plot_grad_mag(x_gt_grad_mag, outdir + "/x_gt")
plot_grad_hist_to_file(x_gt, outdir, "x_gt Gradient Histogram", \
    "x_gt_Gradient_Histogram", 0)
# plot_grad_hist_to_screen(x_gt)

# Function handles for 2D convolution
Ax = lambda x: np.real(ifft2(fft2(x) * otf))
Atx = lambda x: np.real(ifft2(fft2(x) * np.conj(otf)))

# b for all parameter combinations
# for a given image file, attempt to use same b for all runs
# of the program by storing the blurred, noisy values in a file
fname = 'test_images/' + "/" + img_str + "_b.dat"
if os.access(fname, os.R_OK):
    f = open(fname, "rb")
    b = np.load(f)
    f.close()
else:
    b_blur = Ax(x_gt)
    # Make sure there are no negative values
    b_blur[b_blur < 0] = 0

    # Add poisson noise
    b = poisson(b_blur)
    f = open(fname, "wb")
    np.save(f,b)
    f.close()

# compute and plot gradient magnitude of b
b_grad_mag = compute_grad_mag(b)
plot_grad_mag(b_grad_mag, outdir + "/b")
plot_grad_hist_to_file(b, outdir, "b Gradient Histogram", \
    "b_Gradient_Histogram", 0)
# plot_grad_hist_to_screen(b)
# plot_hist_to_screen(b_grad_mag)

# define a threshold for per pixel gradient magnitude
# sparser gradient images will have a higher percentage of pixels
# below the threshold
grad_mag_threshold = 0.5 * np.std(b_grad_mag)
print("grad_mag_threshold: ", grad_mag_threshold)

grad_mag_dirname=None
grad_hist_dirname=None
update_rule_dirname=None

# uncomment to produce per iteration gradient magnitude images
# grad_mag_dirname = outdir + "/grad_mag_dir"

# uncomment to produce per iteration gradient histogram images
# grad_hist_dirname = outdir + "/grad_hist_dir"

# uncomment to produce per iteration x update rule denominator histogram images
# update_rule_dirname = outdir + "/update_rule_dir"

# initial x for all parameter combinations
# for a given image file, attempt to use same initial x for all runs
# of the program by storing the random values in a file
fname = 'test_images/' + "/" + img_str + "_x_init.dat"
if os.access(fname, os.R_OK):
    f = open(fname, "rb")
    x0 = np.load(f)
    f.close()
else:
    x0 = np.random.rand(*x_gt.shape)
    f = open(fname, "wb")
    np.save(f,x0)
    f.close()

lams = [0.00, 0.01, 0.02, 0.03, 0.04, 0.05]
# lams = [0.04]

log_likelihoods = {}
mses = {}
dxyl1norms = {}
gradient_std_devs = {}
pct_grad_mag_thresholds = {}

nIters = 200

for lam in lams:

    results = run_RLTV(nIters, x0, b, Ax, Atx, lam, use_reg_term, \
        grad_mag_threshold, x_gt, grad_mag_dirname, grad_hist_dirname, \
        update_rule_dirname)

    x = results[0]
    log_likelihoods[lam] = results[1]
    mses[lam] = results[2]
    dxyl1norms[lam] = results[3]
    gradient_std_devs[lam] = results[4]
    pct_grad_mag_thresholds[lam] = results[5]

    # plot gradient magnitude for resulting x
    x_grad_mag = compute_grad_mag(x)
    plot_grad_mag(x_grad_mag, outdir + "/x_" + use_reg_term + "_" + \
        "\u03BB=" + str(lam))

    # uncomment to view gradient histogram interactively
    # plot_grad_hist_to_screen(x)

    # convert back to x,y,rgb order and divide out scaling factor
    # for output code below
    x_gt_out = x_gt.transpose(1, 2, 0) / n_photons
    b_out = b.transpose(1, 2, 0) /n_photons
    x_out = x.transpose(1, 2, 0) / n_photons

    psnrx = compute_psnr(x_gt, x, n_photons)
    psnrb = compute_psnr(x_gt, b, n_photons)

    # generate side-by-side comparison of x_gt, b, and x
    fig, axs = plt.subplots(nrows=1, ncols=3, figsize=(10,10))

    x_gt_out = np.clip(x_gt_out, a_min=0.0, a_max=1.0)
    axs[0].imshow(x_gt_out)
    axs[0].axis('off')
    axs[0].set_title("Original image")
    b_out = np.clip(b_out, a_min=0.0, a_max=1.0)
    axs[1].imshow(b_out)
    axs[1].axis('off')
    title = "Blurred image with noise, PSNR=" + \
        str(round(psnrb,2)) + "dB"
    axs[1].set_title('\n'.join(wrap(title, 20)))
    x_out = np.clip(x_out, a_min=0.0, a_max=1.0)
    axs[2].imshow(x_out)
    axs[2].axis('off')
    title = use_reg_term + " filtered image, \u03BB=" + \
        str(lam) + " PSNR=" + str(round(psnrx,2)) + "dB"
    axs[2].set_title('\n'.join(wrap(title, 20)))

    fname = outdir + "/" + use_reg_term + "_" + str(lam) + ".png"
    plt.savefig(fname, bbox_inches='tight')
    plt.close()

    # save full res version of x_out
    fname = outdir + "/" + img_str + "_" + use_reg_term + "_" + str(lam) + ".png"
    io.imsave(fname, (x_out * 255.0).astype(np.uint8))

# save full res version of x_gt_out & b_out
fname = outdir + "/" + img_str + "_gt" + ".png"
io.imsave(fname, (x_gt_out * 255.0).astype(np.uint8))
fname = outdir + "/" + img_str + "_b" + ".png"
io.imsave(fname, (b_out * 255.0).astype(np.uint8))

# plot log likelihoods vs iterations for all parameter combinations
plt.figure()
for lam in lams:
    plt.plot(log_likelihoods[lam], label='\u03BB=' + str(lam))
plt.legend()
plt.title("Log Likelihood, " + use_reg_term)
plt.xlabel('Iteration')
plt.ylabel('Log Likelihood')
fname = outdir + "/" + use_reg_term + "_log_likelihood.png"
plt.savefig(fname, bbox_inches='tight')
plt.close()

# plot MSEs vs iterations for all parameter combinations
plt.figure()
for lam in lams:
    plt.plot(mses[lam], label='\u03BB=' + str(lam))
plt.legend()
plt.title("MSE, " + use_reg_term)
plt.xlabel('Iteration')
plt.ylabel('MSE')
fname = outdir + "/" + use_reg_term + "_mse.png"
plt.savefig(fname, bbox_inches='tight')
plt.close()

# compute ground truth stats to include in plots below
results = compute_ground_truth_stats(x_gt, grad_mag_threshold)
x_gt_dxyl1norm = np.zeros(nIters, float)
x_gt_dxyl1norm[:] = results[0]
x_gt_gradient_std_dev = np.zeros(nIters, float)
x_gt_gradient_std_dev[:] = results[1]
x_gt_pct_grad_mag_threshold = np.zeros(nIters, float)
x_gt_pct_grad_mag_threshold[:] = results[2]

# plot dxyl1norms vs iterations for all parameter combinations
plt.figure()
for lam in lams:
    plt.plot(dxyl1norms[lam], label='\u03BB=' + str(lam))
plt.plot(x_gt_dxyl1norm, label="ground truth", \
    linestyle=(0, (2, 3)))
plt.legend()
plt.title("Dx+Dy L1 Norm, " + use_reg_term)
plt.xlabel('Iteration')
plt.ylabel('L1 Norm')
fname = outdir + "/" + use_reg_term + "_dxyl1norm.png"
plt.savefig(fname, bbox_inches='tight')
plt.close()

# plot gradient std deviations vs iterations for all parameter combinations
plt.figure()
for lam in lams:
    plt.plot(gradient_std_devs[lam], label='\u03BB=' + str(lam))
plt.plot(x_gt_gradient_std_dev, label="ground truth", \
    linestyle=(0, (2, 3)))
plt.legend()
plt.title("Gradient Std Dev, " + use_reg_term)
plt.xlabel('Iteration')
plt.ylabel('Std Dev')
fname = outdir + "/" + use_reg_term + "_grad_std_dev.png"
plt.savefig(fname, bbox_inches='tight')
plt.close()

# plot percent under threshold vs iterations for all parameter combinations
plt.figure()
for lam in lams:
    plt.plot(pct_grad_mag_thresholds[lam], label='\u03BB=' + str(lam))
plt.plot(x_gt_pct_grad_mag_threshold, label="ground truth", \
    linestyle=(0, (2, 3)))
plt.legend()
plt.title("Percent Gradient Magnitude Under Threshold, " + use_reg_term)
plt.xlabel('Iteration')
plt.ylabel('Percent')
fname = outdir + "/" + use_reg_term + "_pct_grad_mag_threshold.png"
plt.savefig(fname, bbox_inches='tight')
plt.close()

# find highest log likelihood, lowest MSE, lowest L1 Norm, lowest gradient
# std dev, and highest percent under threshold across parameter combinations
# ignore the lambda = 0.0 case

# log likelihood
iteration = 0
maxll = -np.Infinity
for lam in lams:
    if lam == 0.0:
        continue
    ll = np.max(log_likelihoods[lam])
    if ll > maxll:
        maxll = ll
        iteration = np.argmax(log_likelihoods[lam])
        maxlam = lam
print("max_log_likelihood: ", '\u03BB=' + str(maxlam), "iteration=", iteration, \
    "ll=", maxll)

# MSE
iteration = 0
minmse = np.Infinity
for lam in lams:
    if lam == 0.0:
        continue
    mse = np.min(mses[lam])
    if mse < minmse:
        minmse = mse
        iteration = np.argmin(mses[lam])
        minlam = lam
print("min_mse: ", '\u03BB=' + str(minlam), "iteration=", iteration, \
    "mse=", minmse)

# L1 Norm
iteration = 0
minl1norm = np.Infinity
for lam in lams:
    if lam == 0.0:
        continue
    dxyl1norm = np.min(dxyl1norms[lam])
    if dxyl1norm < minl1norm:
        minl1norm = dxyl1norm
        iteration = np.argmin(dxyl1norms[lam])
        minlam = lam
print("min_dxyl1norm: ", '\u03BB=' + str(minlam), "iteration=", iteration, \
    "dxyl1norm=", minl1norm)

# gradient standard deviation
iteration = 0
minstddev = np.Infinity
for lam in lams:
    if lam == 0.0:
        continue
    stddev = np.min(gradient_std_devs[lam])
    if stddev < minstddev:
        minstddev = stddev
        iteration = np.argmin(gradient_std_devs[lam])
        minlam = lam
print("min_gradient_std_dev: ", '\u03BB=' + str(minlam), "iteration=", iteration, \
    "stddev=", minstddev)

# percent under threshold
iteration = 0
maxpct = 0.0
for lam in lams:
    if lam == 0.0:
        continue
    pct = np.max(pct_grad_mag_thresholds[lam])
    if pct > maxpct:
        maxpct = pct
        iteration = np.argmax(pct_grad_mag_thresholds[lam])
        maxlam = lam
print("max_pct_under_threshold: ", '\u03BB=' + str(maxlam), "iteration=", iteration, \
    "pct=", maxpct)






